// Copyright (C) 1999 Dominic Letourneau and Andre Charbonneau
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this file.  If not, write to the Free Software Foundation,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef FP_node_info_container_h
#define FP_node_info_container_h

/*
** Includes
*/
#include<iostream.h>
#include<string>
#include<map>
#include<list>
#include "FileParserException.h"

/*
** Defines
*/
#define INT_PARAM 1
#define SHORT_PARAM 2
#define FLOAT_PARAM 3
#define DOUBLE_PARAM 4
#define LONG_PARAM 5 
#define UCHAR_PARAM 6
#define UINT_PARAM 7
#define ULONG_PARAM 8
#define STRING_PARAM 9
#define BOOL_PARAM 10
#define CHAR_PARAM 11
#define SUBNET_PARAM 12
#define EXT_PARAM 13



/**
  * This container class holds the information for a parameter variable of a node.
  */
class param_info
{
  /// The output operator...
  friend ostream& operator<<(ostream&, const param_info& pi);

 public:
  /// The name of the parameter variable.
  string param_name;
  
  /// The value of the parameter variable.
  string param_value;

  /// The type of the parameter variable.
  int param_type;
};



/**
  * This container class holds the information for a input variable of a node.
  */
class input_var_info
{
  /// The output operator...
  friend ostream& operator<<(ostream&, const input_var_info& ii);

 public:
  /// The name of the input variable.
  string var_name;

  /// The name of the source node.
  string source_node_name;

  /// The name of the source output variable.
  string source_var_name;
};



/**
  * This object acts as a container that will hold all the information needed for a node.
  */
class FP_node_info_container
{
  /// The output operator...
  friend ostream& operator<<(ostream&, const FP_node_info_container& nic);

 private:
  /// The name of the node.
  string m_node_name;

  /// The type of the node.
  string m_node_type;

  /// The parameters of the node, with the parameter names as lookup keys.
  map<string,param_info> m_params_map;

  /// The inputs of the node, with the input variable names as lookup keys.
  map<string,input_var_info> m_inputs_map;

 public:
  /*
  ** Constructors / Destructor
  */
  /// Default constructor.
  FP_node_info_container();
  /// Copy constructor.
  FP_node_info_container(const FP_node_info_container& nic);
  /// Destructor.
  virtual ~FP_node_info_container();
  /// = operator
  FP_node_info_container& operator=(const FP_node_info_container& nic);

  /*
  ** Accessors
  */

  /** Sets the node name.
    * @param node_name The node name.
    * @return void
    */
  void set_node_name(const string& node_name);


  /** Gets the node name.
    * @return The name of the node, or "" if the name was not set.
    */
  string get_node_name() const;


  /** Sets the type of the node.
    * @param node_type The type of the node. 
    * @return void
    */
  void set_node_type(const string& node_type);


  /** Gets the type of the node.
    * @return The type of the node, or "" if the type was not set.
    */
  string get_node_type() const;


  /** Adds an input variable to the node.
    * @param var_name The name of the input variable.
    * @param source_node_name The name of the node where the input value will come from.
    * @param source_var_name The name of the output variable where the input value will come from.
    * @return void
    * @exception FileParserException
    */
  void add_input_var(const string& var_name, const string& source_node_name, const string& source_var_name) throw(FileParserException);


  /** Gets the info about an input variable.
    * @param input_name The name of the input variable we are looking for.
    * @return A input_var_info structure containing the information about the input variable.
    * @exception FileParserException
    */
  input_var_info get_input_var_info(const string& input_name) const throw(FileParserException);


  /** Adds a parameter variable to the node.
    * @param param_name The name of the parameter variable.
    * @param param_value The value of the parameter (as a string).
    * @param param_type The type of the parameter variable. Must be in {NUMERIC_PARAM,STRING_PARAM,BOOL_PARAM,BOOL_PARAM}
    * @return void
    * @exception FileParserException
    */
  void add_param_var(const string& param_name, const string& param_value, const int param_type) throw(FileParserException);

  /** Modifies a parameter variable to the node.
    * @param param_name The name of the parameter variable to modify.
    * @param param_value The new value of the parameter (as a string).
    * @param param_type The new type of the parameter variable. Must be in {NUMERIC_PARAM,STRING_PARAM,BOOL_PARAM,BOOL_PARAM}
    * @return void
    * @exception FileParserException
    */
  void modify_param_var(const string& param_name, const string& param_value, const int param_type) throw(FileParserException);

  /** Gets the information about a parameter variable of the node.
    * @param param_name The name of the parameter variable.
    * @return A param_info structure containing the information of the parameter variable.
    * @exception FileParserException
    */
  param_info get_param_info(const string& param_name) const throw(FileParserException);
  

  /** Gets the names of the input variables of the node.
    * @return A list of input variable names as string objects.
    */
  list<string> get_input_var_names() const;


  /** Gets the names of the parameter variables of the node.
    * @return A list of parameter variable names as string objects.
    */
  list<string> get_param_var_names() const;

  /** Deletes all parameters from the node.
    */
  void reset_params();


};
#endif
 
